/*
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * Copyright (C) 1995 - 2000 by Ralf Baechle
 */
#include <linux/config.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/types.h>
#include <linux/ptrace.h>
#include <linux/mman.h>
#include <linux/mm.h>
#include <linux/smp.h>
#include <linux/smp_lock.h>
#include <linux/version.h>
#include <linux/plog.h>
#include <linux/kernel_stat.h>

#include <asm/branch.h>
#include <asm/hardirq.h>
#include <asm/pgalloc.h>
#include <asm/mmu_context.h>
#include <asm/softirq.h>
#include <asm/system.h>
#include <asm/uaccess.h>

extern void show_trace_task(struct task_struct *);

#define development_version (LINUX_VERSION_CODE & 0x100)

/*
 * Macro for exception fixup code to access integer registers.
 */
#define dpf_reg(r) (regs->regs[r])

extern spinlock_t timerlist_lock;

/*
 * Unlock any spinlocks which will prevent us from getting the
 * message out (timerlist_lock is acquired through the
 * console unblank code)
 */
void bust_spinlocks(int yes)
{
	spin_lock_init(&timerlist_lock);
	if (yes) {
		oops_in_progress = 1;
#ifdef CONFIG_SMP
		/* Many serial drivers do __global_cli() */
		global_irq_lock = SPIN_LOCK_UNLOCKED;
#endif
	} else {
		int loglevel_save = console_loglevel;
#ifdef CONFIG_VT
		unblank_screen();
#endif
		oops_in_progress = 0;
		/*
		 * OK, the message is on the console.  Now we call printk()
		 * without oops_in_progress set so that printk will give klogd
		 * a poke.  Hold onto your hats...
		 */
		console_loglevel = 15;		/* NMI oopser may have shut the console up */
		printk(" ");
		console_loglevel = loglevel_save;
	}
}

void _log_user_fault(const char *fn, int sig, struct pt_regs *regs,
		    struct task_struct *tsk)
{
	printk("%s: sending signal %d to %s(%d)\n", fn, sig, tsk->comm,
	       	tsk->pid);
	show_regs(regs);
	show_trace_task(tsk);
}

/*
 * This routine handles page faults.  It determines the address,
 * and the problem, and then passes it off to one of the appropriate
 * routines.
 */
#ifdef CONFIG_4KSTACK
extern unsigned int kernelsp;
#endif

#ifdef CONFIG_TIVO_DEVEL

#include <asm/inst.h>

/* magic_mempool_modify
 * This routine is a dummy routine which is modified by
 * handle_magic_mempool_access to insert the offending
 * instruction.
 */

int magic_mempool_modify(unsigned int a0, unsigned int a1) {
	int ret;
	__asm__ __volatile__ (
		"1:\tmove $0, $0\n\t"
		"move %0,$5\n"
		"2:\n\t"
		".section .fixup,\"ax\"\n"
		"3:\tli %0,%1\n\t"
		"j 2b\n\t"
		".previous\n\t"
		".section __ex_table,\"a\"\n\t"
		".word 1b + 0x20000000,3b\n\t"
		".previous"
		:"=r" (ret)
		: "i" (-EFAULT));
	return ret;
}

/*
 * test_emit_backtrace
 * This tests for the condition of over-writing a freed block by
 * seeing if the address written to contains DEADBEEF.  If it
 * does, it emits the backtrace.
 * 
 */

inline test_emit_backtrace(struct pt_regs* regs, unsigned long address,
			   unsigned int a1) {
	unsigned int val;
	address &= ~(0x3); /* Word align the address */
	address |= 0x10000000; /* Move to non magic 0x5xxxxxxx */
	__get_user(val, (unsigned int*)address);
	if(val != 0xdeadbeef) {
		return;
	}
	/* We're over-writing deadbeef, this is broken in all
	 * cases since currently mempools write to 0x5....,
	 * if in the future we suspect bugs in the mempool code
	 * the following conditional should be added:
	 * if(a1 != 0xaaaaaaaa) ....
	 */
	printk("\nDEADBEEF_OVERWRITE of value 0x%08X at addr 0x%p"
               " on pid %d\n",
               a1,
               address,
               current->pid);
	show_trace_task(current);	
	
}

unsigned int start_magic_addrs=0x40000000;
unsigned int end_magic_addrs=  0x4fffffff;

/*
 * handle_magic_mempool_access
 * This function is called when a fault has occured to our magic region.
 * It handles this fault by re-executing the offending instruction with
 * the modified address.
 * 
 */

int handle_magic_mempool_access(struct pt_regs* regs, unsigned long write,
				unsigned long address) {
	/* It's a magic mempool address, fun for everyone! */
	/* A "magic" mempool has two user addresses, one is a 
	 * bogus mapping in the 0x4....... address space, and one
	 * is a valid mapping in the 0x5....... address space.
	 * The writes to 0x4 push us up into the kernel so we
	 * can write out and monitor the activity.
	 */

	/* Emulate instruction */
	unsigned long pc;
	union mips_instruction insn, insnOrig;
	unsigned int a0, a1;
	unsigned int* rewrite_addr=(unsigned int*)(&magic_mempool_modify + 0x20000000); /* Uncached addr */
	int (*mmpool_uncached)(unsigned int, unsigned int);
	static unsigned int last_ll_context_count;
#if 0
	static int mmagicaccess=0;
#endif
	mmpool_uncached=rewrite_addr;
	pc = regs->cp0_epc + ((regs->cp0_cause & CAUSEF_BD) ? 4 : 0);
	__get_user(insn.word, (unsigned int *)pc);
	insnOrig.word=insn.word;
	a0=regs->regs[insn.i_format.rs]; /* base */
	a1=regs->regs[insn.i_format.rt]; /* rt   */
	if(write) {
		test_emit_backtrace(regs, address, a1);
	}

	a0 +=0x10000000; /* Move base into correct 0x5xxxxxxx area */
	insn.i_format.rs=4; /* Change base to a0 */
	insn.i_format.rt=5; /* Change rt to a1 */

	switch (insn.i_format.opcode) {
		/* This code makes ll/sc work on our "magic" areas.
		 * It does this by storing the context switch counter,
		 * if the context switch counter increments since our
		 * last ll, we let it break.  Otherwise, we re-execute
		 * the ll instruction so our sc will be cool.
		 */
	case ll_op:
		/* Save our context count */
		last_ll_context_count=kstat.context_swtch;
		break;
	case sc_op:
		if(last_ll_context_count == kstat.context_swtch) {
			/* Set llbit flag */
			/* We do this the cheesiest way we know how...,
			   simply re-execute the ll :\
			*/
			union mips_instruction ll_insn;
			ll_insn.word=insn.word;
			ll_insn.i_format.opcode=ll_op;
			rewrite_addr[0]=ll_insn.word;
			mmpool_uncached(a0,a1);
			/* a1 might be bogus, so lets reload to be sure */
			a1=regs->regs[insnOrig.i_format.rt];
			/* We'll fall through here, and run our sc as normal */
		} else {
			/* Don't need to do anything here, the sc will break
			   down below, and we want it to.
			*/
		}
		break;
	}

	rewrite_addr[0]=insn.word;
	a1=mmpool_uncached(a0, a1);
	compute_return_epc(regs);
	regs->regs[insnOrig.i_format.rt]=a1;
#if 0
	if((mmagicaccess++ % 10000) == 0) {
		printk("magic access %d, write %d", mmagicaccess, write);
		printk(" BD %d, insn %08x, addr %08x, a0 %08x, pc %08x\n", regs->cp0_cause & CAUSEF_BD ? 1:0, insn.word, address, a0, pc);
		extern void dump_tlb_all(void);
		dump_tlb_all();
	}
  
#endif
	return;
}

#endif

asmlinkage void do_page_fault(struct pt_regs *regs, unsigned long write,
			      unsigned long address)
{
	struct vm_area_struct * vma;
	struct task_struct *tsk = current;
	struct mm_struct *mm = tsk->mm;
	unsigned long fixup;
	siginfo_t info;
	int ret;

#ifdef CONFIG_TIVO_DEVEL
	if(address >= start_magic_addrs && address <= end_magic_addrs) {
		handle_magic_mempool_access(regs, write, address);
		return;
	}

#endif
	
	plog_generic_event3(PLOG_TYPE_PFE, (int)address, (int)regs->cp0_epc,
			(int)regs->regs[31]);

	/*
	 * We fault-in kernel-space virtual memory on-demand. The
	 * 'reference' page table is init_mm.pgd.
	 *
	 * NOTE! We MUST NOT take any locks for this case. We may
	 * be in an interrupt or a critical region, and should
	 * only copy the information from the master page table,
	 * nothing more.
	 */
	if (address >= VMALLOC_START)
		goto vmalloc_fault;

	info.si_code = SEGV_MAPERR;
	/*
	 * If we're in an interrupt or have no user
	 * context, we must not take the fault..
	 */
	if (in_interrupt() || !mm)
		goto no_context;
#if 0
#ifdef CONFIG_4KSTACK
	printk("[%s:%d:%08lx:%ld:%08lx:%08lx]\n", current->comm, current->pid,
	       address, write, regs->cp0_epc, kernelsp);
#else
	printk("[%s:%d:%08lx:%ld:%08lx]\n", current->comm, current->pid,
	       address, write, regs->cp0_epc);
#endif
#endif
	down_read(&mm->mmap_sem);
	vma = find_vma(mm, address);
	if (!vma)
		goto bad_area;
	if (vma->vm_start <= address)
		goto good_area;
	if (!(vma->vm_flags & VM_GROWSDOWN))
		goto bad_area;
	if (expand_stack(vma, address))
		goto bad_area;
/*
 * Ok, we have a good vm_area for this memory access, so
 * we can handle it..
 */
good_area:
	info.si_code = SEGV_ACCERR;

	if (write) {
		if (!(vma->vm_flags & VM_WRITE))
			goto bad_area;
	} else {
		if (!(vma->vm_flags & (VM_READ | VM_EXEC)))
			goto bad_area;
	}

survive:
	/*
	 * If for any reason at all we couldn't handle the fault,
	 * make sure we exit gracefully rather than endlessly redo
	 * the fault.
	 */
	ret = handle_mm_fault(mm, vma, address, write, tsk->rt_priority);
	plog_generic_event(PLOG_TYPE_PFL, ret);
	switch (ret) {
	case 1:
		tsk->min_flt++;
		break;
	case 2:
		tsk->maj_flt++;
		break;
	case 0:
		goto do_sigbus;
	default:
		goto out_of_memory;
	}

	up_read(&mm->mmap_sem);
	return;

/*
 * Something tried to access memory that isn't in our memory map..
 * Fix it, but check if it's kernel or user first..
 */
bad_area:
	up_read(&mm->mmap_sem);

bad_area_nosemaphore:
	plog_generic_event(PLOG_TYPE_PFL, -2);
	/* User mode accesses just cause a SIGSEGV */
	if (user_mode(regs)) {
		tsk->thread.cp0_badvaddr = address;
		tsk->thread.error_code = write;
#if 1
		printk("Illegal %s at %08lx\n", write ? "write" : "read",
		       	address);
		_log_user_fault(__FUNCTION__ " #2", SIGSEGV, regs, tsk);
#endif
		info.si_signo = SIGSEGV;
		info.si_errno = 0;
		/* info.si_code has been set above */
		info.si_addr = (void *) address;
		force_sig_info(SIGSEGV, &info, tsk);
		return;
	}

no_context:
	/* Are we prepared to handle this kernel fault?  */
	fixup = search_exception_table(exception_epc(regs));
	if (fixup) {
		long new_epc;

		tsk->thread.cp0_baduaddr = address;
		new_epc = fixup_exception(dpf_reg, fixup, regs->cp0_epc);
		if (development_version)
			printk(KERN_DEBUG "%s: Exception at [<%lx>] (%lx)\n",
			       tsk->comm, regs->cp0_epc, new_epc);
		regs->cp0_epc = new_epc;
		return;
	}

	/*
	 * Oops. The kernel tried to access some bad page. We'll have to
	 * terminate things with extreme prejudice.
	 */
	printk(KERN_ALERT "Unable to handle kernel paging request at virtual "
	       "address %08lx, epc == %08lx, ra == %08lx\n",
	       address, regs->cp0_epc, regs->regs[31]);
	die("Oops", regs);
	/* Game over.  */

/*
 * We ran out of memory, or some other thing happened to us that made
 * us unable to handle the page fault gracefully.
 */
out_of_memory:
	if (tsk->pid == 1) {
		yield();
		goto survive;
	}
	up_read(&mm->mmap_sem);
	printk(KERN_NOTICE "VM: killing process %s\n", tsk->comm);
	if (user_mode(regs))
		do_exit(SIGKILL);
	goto no_context;

do_sigbus:
	up_read(&mm->mmap_sem);

	/*
	 * Send a sigbus, regardless of whether we were in kernel
	 * or user mode.
	 */
	tsk->thread.cp0_badvaddr = address;
#if 1
	if (user_mode(regs))
		_log_user_fault(__FUNCTION__ " #3", SIGBUS, regs, tsk);
#endif
	info.si_signo = SIGBUS;
	info.si_errno = 0;
	info.si_code = BUS_ADRERR;
	info.si_addr = (void *) address;
	force_sig_info(SIGBUS, &info, tsk);

	/* Kernel mode? Handle exceptions or die */
	if (!user_mode(regs))
		goto no_context;

	return;

vmalloc_fault:
	{
		/*
		 * Synchronize this task's top level page-table
		 * with the 'reference' page table.
		 */
		int offset = pgd_index(address);
		pgd_t *pgd, *pgd_k;
		pmd_t *pmd, *pmd_k;

		pgd = tsk->active_mm->pgd + offset;
		pgd_k = init_mm.pgd + offset;

		if (!pgd_present(*pgd)) {
			if (!pgd_present(*pgd_k))
				goto bad_area_nosemaphore;
			set_pgd(pgd, *pgd_k);
			return;
		}

		pmd = pmd_offset(pgd, address);
		pmd_k = pmd_offset(pgd_k, address);

		if (pmd_present(*pmd) || !pmd_present(*pmd_k))
			goto bad_area_nosemaphore;
		set_pmd(pmd, *pmd_k);
	}
}
