/*
 * SATA drive diagonsis hooks
 *
 * Copyright (C) 2006 TiVo Inc. All Rights Reserved.
 */

#include <linux/module.h>
#include <linux/proc_fs.h>
#include <asm/uaccess.h>
#include <linux/ide.h>
#include <linux/pci.h>
#include <asm/bcm7038.h>

#include "sata-diag.h"

#define NULLENTRY	{ NULL, NULL, -1, 0, NULL, NULL, -1 }

#if 0
#undef SATA_MEM_BASE
#define SATA_MEM_BASE	PCI_EXPANSION_PHYS_MEM_WIN0_BASE
#endif

#define DIAG_IN_WORD	mmio_inl
#define DIAG_OUT_WORD	mmio_outl

static int diag_start_read
	(char *page, char **start, off_t off, int count, int *eof, void *data);
static int diag_start_write
	(struct file *file, const char *buffer, unsigned long count, void *data);

static int mmio_reg_read
	(char *page, char **start, off_t off, int count, int *eof, void *data);
static int mmio_reg_write
	(struct file *file, const char *buffer, unsigned long count, void *data);

static int test_reg_read
	(char *page, char **start, off_t off, int count, int *eof, void *data);
static int test_reg_write
	(struct file *file, const char *buffer, unsigned long count, void *data);

/* Proc entries under SATA-diagnosis */
sata_proc_entry_t sata_diag_reg_entries[] = {
	{
		"diag_start_port0",
		NULL,
		0,
		S_IFREG|S_IRUSR|S_IRGRP|S_IROTH,
		diag_start_read,
		diag_start_write,
		0
	},
	{
		"diag_start_port1",
		NULL,
		0,
		S_IFREG|S_IRUSR|S_IRGRP|S_IROTH,
		diag_start_read,
		diag_start_write,
		1
	},
	{
		"mdio_access_reg",
		NULL,
		MDIO_ACCESS_REG,
		S_IFREG|S_IRUSR|S_IRGRP|S_IROTH,
		mmio_reg_read,
		mmio_reg_write,
		0
	},
	{
		"mmio_sata_status_port0",
		NULL,
		MMIO_SATA_STATUS,
		S_IFREG|S_IRUSR|S_IRGRP|S_IROTH,
		mmio_reg_read,
		mmio_reg_write,
		0
	},
	{
		"mmio_sata_status_port1",
		NULL,
		MMIO_SATA_STATUS,
		S_IFREG|S_IRUSR|S_IRGRP|S_IROTH,
		mmio_reg_read,
		mmio_reg_write,
		1
	},
	{
		"mmio_sata_error_port0",
		NULL,
		MMIO_SATA_ERROR,
		S_IFREG|S_IRUSR|S_IRGRP|S_IROTH,
		mmio_reg_read,
		mmio_reg_write,
		0
	},
	{
		"mmio_sata_error_port1",
		NULL,
		MMIO_SATA_ERROR,
		S_IFREG|S_IRUSR|S_IRGRP|S_IROTH,
		mmio_reg_read,
		mmio_reg_write,
		1
	},
	{
		"mmio_sata_control_port0",
		NULL,
		MMIO_SATA_CONTROL,
		S_IFREG|S_IRUSR|S_IRGRP|S_IROTH,
		mmio_reg_read,
		mmio_reg_write,
		0
	},
	{
		"mmio_sata_control_port1",
		NULL,
		MMIO_SATA_CONTROL,
		S_IFREG|S_IRUSR|S_IRGRP|S_IROTH,
		mmio_reg_read,
		mmio_reg_write,
		1
	},
	{
		"mmio_sata_mask_port0",
		NULL,
		MMIO_SATA_MASK,
		S_IFREG|S_IRUSR|S_IRGRP|S_IROTH,
		mmio_reg_read,
		mmio_reg_write,
		0
	},
	{
		"mmio_sata_mask_port1",
		NULL,
		MMIO_SATA_MASK,
		S_IFREG|S_IRUSR|S_IRGRP|S_IROTH,
		mmio_reg_read,
		mmio_reg_write,
		1
	},
	{
		"mmio_sata_control2_port0",
		NULL,
		MMIO_SATA_CONTROL2,
		S_IFREG|S_IRUSR|S_IRGRP|S_IROTH,
		mmio_reg_read,
		mmio_reg_write,
		0
	},
	{
		"mmio_sata_control2_port1",
		NULL,
		MMIO_SATA_CONTROL2,
		S_IFREG|S_IRUSR|S_IRGRP|S_IROTH,
		mmio_reg_read,
		mmio_reg_write,
		1
	},
	{
		"sata_BIST_FIS_PORT0_ctrl",
		NULL,
		SATA_BIST_FIST_CTRL,
		S_IFREG|S_IRUSR|S_IRGRP|S_IROTH,
		mmio_reg_read,
		mmio_reg_write,
		0
	},
	{
		"sata_BIST_FIS_PORT1_ctrl",
		NULL,
		SATA_BIST_FIST_CTRL,
		S_IFREG|S_IRUSR|S_IRGRP|S_IROTH,
		mmio_reg_read,
		mmio_reg_write,
		1
	},
	{
		"test_mem_space",
		NULL,
		0,
		S_IFREG|S_IRUSR|S_IRGRP|S_IROTH,
		test_reg_read,
		test_reg_write,
		1
	},
	NULLENTRY
};

struct proc_dir_entry * sata_diag_proc_entry = NULL;

const struct pci_dev *sata_dev = NULL;
const struct pci_dev *bcm7038_dev = NULL;

void dum_pci_device(void)
{

	sata_dev = pci_find_device (PCI_VENDOR_ID_SERVERWORKS,
			PCI_DEVICE_ID_SERVERWORKS_BCM7038,
			NULL);

	if(sata_dev == NULL)
	{
		printk("wrong device\n");
	}

	bcm7038_dev = pci_find_device (PCI_VENDOR_ID_BROADCOM,
			0x7042,
			NULL);

	if(bcm7038_dev == NULL)
	{
		printk("wrong device 7038\n");
	}

	if(sata_dev)
	printk("Proc operations = %p \n", sata_dev->bus->ops->read_byte);

}

//#define PROC_DEBUG	1

#ifdef PROC_DEBUG

static void print_pci_config()
{
	int i;

	for (i=0; i< 0xf; i++)
		printk("Reg addr = %x, value = %x\n",
				i,
				IN_WORD(0x200+i));

}
#endif

static int diag_start_read(char * buffer, char ** start, off_t offset,
				int size, int * eof, void * data)
{
	int port, len = 0, status;
	sata_proc_entry_t *ent;

#ifdef	PROC_DEBUG

	print_pci_config();

	*eof = 1;
	return 0;
#endif

	ent = (sata_proc_entry_t *)data;

	port = (ent->private? PORT_OFFSET: 0);

	//printk("diag_start_read: reading offset=0x%x ", MMIO_SATA_STATUS+port);
	status = DIAG_IN_WORD(MMIO_SATA_STATUS+port);
	//printk("value=0x%x\n", status);

	if(status & PHY_IN_BIST)
	{
		len += sprintf(buffer,"%s\n","BIST started");
	}
	else
	{
		len += sprintf(buffer,"%s\n","BIST disabled");
	}

	*eof = 1;
	return len;
}

static int diag_start_write(struct file * filp, const char * buffer,
				 unsigned long count, void * data)
{
	int val, status, port;
	int num;
	char buff[11];
	sata_proc_entry_t *ent;

	if (!capable(CAP_SYS_ADMIN) )
		return -EACCES;

	num = (count < 10) ? count : 10;
	copy_from_user(buff, buffer, num);
	buff[num] = '\0';

	val = simple_strtoul(buff, 0, 0);

	ent = (sata_proc_entry_t *)data;

	port = (ent->private? PORT_OFFSET: 0);

	//printk("diag_start_write: reading offset=0x%x ", MMIO_SATA_STATUS+port);
	status = DIAG_IN_WORD(MMIO_SATA_STATUS+port);
	//printk("value=0x%x\n", status);

	if((status & PHY_IN_BIST) && (val == 0))
	{
		//printk("diag_start_write: reading offset=0x%x ", MMIO_SATA_CONTROL+port);

		status = DIAG_IN_WORD(MMIO_SATA_CONTROL+port);
		//printk("value=0x%x\n", status);

		/* The first 4 bits */
		status &= 0xFFFFFFF0;

		status |= 0x1;

		//printk("diag_start_write: writing offset=0x%x value=0x%x\n",
						//MMIO_SATA_CONTROL+port, status);

		DIAG_OUT_WORD(MMIO_SATA_CONTROL+port, status);

		/* TODO: Need to add some code after input from broadcom */
		/* Disable BIST */
	}
	else if(!(status & PHY_IN_BIST) && (val != 0))
	{
		/* Do nothing.. expect the script to do it all */
		/* Enable BIST */
	}

	return count;
}

static int mmio_reg_read(char * buffer, char ** start, off_t offset,
				int size, int * eof, void * data)
{
	int port, len = 0, status;
	sata_proc_entry_t *ent;

	ent = (sata_proc_entry_t *)data;

	port = (ent->private? PORT_OFFSET: 0);

	//printk("mmio_reg_read: offset=0x%x ", ent->identity+port);

	status = DIAG_IN_WORD(ent->identity+port);

	//printk("value=0x%x\n", status);

	len += sprintf(buffer,"0x%x\n",status);

	*eof=1;
	return len;
}

int test_interrupt;

static int mmio_reg_write(struct file * filp, const char * buffer,
				 unsigned long count, void * data)
{
	int val, port;
	int num;
	char buff[11];
	sata_proc_entry_t *ent;

	if (!capable(CAP_SYS_ADMIN) )
		return -EACCES;

	num = (count < 10) ? count : 10;
	copy_from_user(buff, buffer, num);
	buff[num] = '\0';

	val = simple_strtoul(buff, 0, 0);

	ent = (sata_proc_entry_t *)data;

	port = (ent->private? PORT_OFFSET: 0);

	//printk("mmio_reg_write: offset=0x%x value=0x%x\n", ent->identity+port, val);

	DIAG_OUT_WORD(ent->identity+port, val);

	if(ent->identity+port == 0x188)
	{
		//test_interrupt = 1;
	}

	return count;
}

static int test_reg_read(char * buffer, char ** start, off_t offset,
				int size, int * eof, void * data)
{
	dum_pci_device();
	*eof = 1;
	return 0;
}

static int sata_pci_read_config_dword( struct pci_dev *dev,
				int where, u32 *val)
{
    int slot_num, func_num;
    u32 cfc,cf8;
    volatile u32 *cfg_index_reg;
    volatile u32 *cfg_data_reg;

	cfg_index_reg	=   (volatile u32 *)PCI_BRIDGE_SATA_CFG_INDEX;
	cfg_data_reg	=   (volatile u32 *)PCI_BRIDGE_SATA_CFG_DATA; 

    slot_num = PCI_SLOT(dev->devfn);	
    func_num = PCI_FUNC(dev->devfn);
    cf8 = PCI_ENABLE + PCI_IDSEL(slot_num) + PCI_FNCSEL(func_num) + where;

	//printk("index reg value = %x\n", cf8);
    *cfg_index_reg = cf8;
    cfc = *cfg_data_reg;
    *val = cfc;
    
    return PCIBIOS_SUCCESSFUL;
}

static int bcm7038_pci_read_config_dword( struct pci_dev *dev,
				int where, u32 *val)
{
    int slot_num, func_num;
    u32 cf8;
    volatile u32 *cfg_index_reg;
    volatile u32 *cfg_data_reg;

	cfg_index_reg	=   (volatile u32 *)MIPS_PCI_XCFG_INDEX;
	cfg_data_reg	=   (volatile u32 *)MIPS_PCI_XCFG_DATA; 

    slot_num = PCI_SLOT(dev->devfn);	
    func_num = PCI_FUNC(dev->devfn);
    cf8 = PCI_ENABLE + PCI_IDSEL(slot_num) + PCI_FNCSEL(func_num) + where;

	//printk("index reg value = %p\n", (BCM_PCI_VENDOR_SELF+where));
#if 0
    *cfg_index_reg = cf8;
    cfc = *cfg_data_reg;
    *val = cfc;
#else
    cfg_index_reg = (volatile u32 *)(((unsigned int )BCM_PCI_VENDOR_SELF)+where);
    *val = *(cfg_index_reg);
#endif
    
    return PCIBIOS_SUCCESSFUL;
}

static int read_bcm7038_pci_space(int offset)
{
		int val = 0;

	if(bcm7038_dev == NULL)
	{
		dum_pci_device();
	}

	if(bcm7038_dev == NULL)
	{
		printk("Unable to find bcm7038 pci device\n");
		return 0;
	}

	//printk("access offset = %x\n", offset);
//	val = _tivo_ide_in_word(offset);
	bcm7038_pci_read_config_dword(bcm7038_dev, offset, &val);

	return val;
}

static int read_sata_pci_space(int offset)
{
		int val = 0;

	if(sata_dev == NULL)
	{
		dum_pci_device();
	}

	if(sata_dev == NULL)
	{
		printk("Unable to find pci device\n");
		return 0;
	}

	printk("access offset = %x\n", offset);
//	val = _tivo_ide_in_word(offset);
	sata_pci_read_config_dword(sata_dev, offset, &val);
	//bcm7038_pci_read_config_dword(bcm7038_dev, offset, &val);

	return val;
}

void dump_bcm7038_cfg_value()
{
	printk("Vendor ID 		= 0x%x \n",read_bcm7038_pci_space(0x0));
	printk("Rev ID Class 		= 0x%x \n",read_bcm7038_pci_space(0x8));
	printk("Cacheline size 		= 0x%x \n",read_bcm7038_pci_space(0xc));
	printk("GISB BASE W		= 0x%x \n",read_bcm7038_pci_space(0x10));
	printk("MEM BASE W0		= 0x%x \n",read_bcm7038_pci_space(0x14));
	printk("MEM BASE W1		= 0x%x \n",read_bcm7038_pci_space(0x18));
	printk("MEM BASE W2		= 0x%x \n",read_bcm7038_pci_space(0x1c));

	printk("Sub System ID 		= 0x%x \n",read_bcm7038_pci_space(0x2c));
	printk("Interrupt Line 		= 0x%x \n",read_bcm7038_pci_space(0x3c));
	printk("CPU 2 PCI MEM WIN0	= 0x%x \n",read_bcm7038_pci_space(0x50));
	printk("CPU 2 PCI MEM WIN1	= 0x%x \n",read_bcm7038_pci_space(0x54));
	printk("CPU 2 PCI MEM WIN2	= 0x%x \n",read_bcm7038_pci_space(0x58));
	printk("CPU 2 PCI MEM WIN3	= 0x%x \n",read_bcm7038_pci_space(0x5c));
	printk("CPU 2 PCI IO  WIN0	= 0x%x \n",read_bcm7038_pci_space(0x60));
	printk("CPU 2 PCI IO  WIN1	= 0x%x \n",read_bcm7038_pci_space(0x64));
	printk("CPU 2 PCI IO  WIN2	= 0x%x \n",read_bcm7038_pci_space(0x68));
	printk("SDRAM ENDIAN CTRL	= 0x%x \n",read_bcm7038_pci_space(0x6c));
}

void dump_sata_cfg_value()
{
    volatile u32 *cfg_index_reg;

	cfg_index_reg	=   (volatile u32 *)PCI_BRIDGE_SATA_CFG_INDEX;

	printk("Vendor ID 		= 0x%x \n",read_sata_pci_space(0x0));
	printk("Device ID 		= 0x%x \n",read_sata_pci_space(0x2));
	printk("PCI Status 		= 0x%x \n",read_sata_pci_space(0x6));
	printk("Rev ID Class 		= 0x%x \n",read_sata_pci_space(0x8));
	printk("Cacheline size 		= 0x%x \n",read_sata_pci_space(0xc));
	printk("Master Latecy 		= 0x%x \n",read_sata_pci_space(0xd));
	printk("Header Type 		= 0x%x \n",read_sata_pci_space(0xe));
	printk("Bist 			= 0x%x \n",read_sata_pci_space(0xf));
	printk("CFG PCS0 		= 0x%x \n",read_sata_pci_space(0x10));
	printk("CFG PCS1 		= 0x%x \n",read_sata_pci_space(0x14));
	printk("CFG SCS0 		= 0x%x \n",read_sata_pci_space(0x18));
	printk("CFG SCS1 		= 0x%x \n",read_sata_pci_space(0x1c));
	printk("Bus Master Base Addr 	= 0x%x \n",read_sata_pci_space(0x20));
	printk("Mimo base addr 		= 0x%x \n",read_sata_pci_space(0x24));
	printk("Sub Vendor ID 		= 0x%x \n",read_sata_pci_space(0x2c));
	printk("Sub device ID 		= 0x%x \n",read_sata_pci_space(0x2e));
	printk("Interrupt Line 		= 0x%x \n",read_sata_pci_space(0x3c));
	printk("Interrupt Pin 		= 0x%x \n",read_sata_pci_space(0x3d));
	printk("Minimum Grant 		= 0x%x \n",read_sata_pci_space(0x3e));
	printk("Maximum Latency 	= 0x%x \n",read_sata_pci_space(0x3f));

	cfg_index_reg -= 1;

	printk("PCI Bridge Ctrl			=0x%x	\n",*(cfg_index_reg));
	cfg_index_reg += 3;
	printk("PCI Bridge SLV MEM WIN BASE	=0x%x	\n",*(cfg_index_reg));
	cfg_index_reg += 1;
	printk("PCI Bridge CPU MEM WIN BASE	=0x%x	\n",*(cfg_index_reg));
	cfg_index_reg += 1;
	printk("PCI Bridge CPU IO  WIN BASE	=0x%x	\n",*(cfg_index_reg));
}

static int test_reg_write(struct file * filp, const char * buffer,
				 unsigned long count, void * data)
{
	int num, address;
	char buff[11];

	num = (count < 10) ? count : 10;
	copy_from_user(buff, buffer, num);
	buff[num] = '\0';

	address = simple_strtoul(buff, 0, 0);

#if 0
	printk("offset = %x, value = %x\n",
			address, read_sata_pci_space(address));
#else
	printk("PCI Sata Configuration ...\n");

	dump_sata_cfg_value();
	printk("PCI Configuration ..\n");
	dump_bcm7038_cfg_value();
#if 0
    	printk(" SATA STATUS = 0x%x\n",
			*((volatile u32 *)(((unsigned int )BCM_PCI_VENDOR_SELF)+0x520440)));
#endif
#endif

	return count;
}

void create_proc_diag_registry_interfaces(struct proc_dir_entry *parent)
{
	struct proc_dir_entry *ent;
	sata_proc_entry_t *entries = sata_diag_reg_entries;

	if(!parent)
		return;

	while(entries->name != NULL)
	{
		ent = create_proc_entry(entries->name, entries->mode, parent);
		if(!ent)
		{
			printk("Failed to create proc interface for %s\n",
					entries->name);
			return;
		}
		entries->proc_entry = ent;
		ent->nlink = 1;
		ent->data = (void *)entries;
		ent->read_proc = entries->read_proc;
		ent->write_proc = entries->write_proc;
		entries++;
	}
}

static int sata_diag_proc_init(void)
{
	if ( sata_diag_proc_entry )
		return 1; 

	sata_diag_proc_entry  = create_proc_entry("sata_diag", S_IFDIR, NULL);

	if (sata_diag_proc_entry == NULL) {
		printk("Unable to create proc entry for sata_diag!\n");
		return 1;
	}

	create_proc_diag_registry_interfaces(sata_diag_proc_entry);
	return 0;
}

static int __init sata_diag_init(void)
{
	sata_diag_proc_init();
	return 0;
}
__initcall(sata_diag_init);

static void sata_diag_proc_cleanup(void)
{
	if (! sata_diag_proc_entry)
		return;
	remove_proc_entry("sata-diag", sata_diag_proc_entry);
}

