#ifndef _SGAT_DIO_H
#define _SGAT_DIO_H
#ifdef __KERNEL__
#include <linux/uio.h> /* for struct iovec */
#else 
#include <sys/uio.h> /* for struct iovec */
#endif /* __KERNEL__ */

#define SZ_SGAT_DIO_HDR sizeof(Sgat_dio_hdr_t)
#define SZ_SGAT_IOVEC sizeof(struct iovec)
#define SZ_SGAT_REQ sizeof(Sgat_dio_request)

#define SGAT_DIO_DXFER_TO_DEV -1   
#define SGAT_DIO_DXFER_FROM_DEV -2  
#define SGAT_DIO_DXFER_TO_FROM_DEV -3 
///////////////////////////////////////////////////////////////////////////////
//
//Data Definitions
//
//////////////////////////////////////////////////////////////////////////////

typedef struct sgat_dio_hdr
{
    int dxfer_dir;		/* [i] data transfer direction  */
    unsigned short iovec_count; /* [i] 0 implies no scatter gather */
    unsigned int dxfer_len;     /* [i] byte count of data transfer */
    unsigned long timeout;       /* [i] MAX_UINT->no timeout (unit: millisec) */
    int status;			/* wiil be copied back to user space*/
    unsigned short xferId;    
    struct iovec  * sglp ;      /* [i] points to user scatter gather list */
} Sgat_dio_hdr_t;  

#ifdef __KERNEL__

typedef struct sgat_dio_request  /* holding area for scsi scatter gather info */
{
    unsigned long timeout;
    unsigned short xferId;
    unsigned short sclp_count;  /* Count of kernel scatter-gather pieces */
    int dxfer_dir;		/* [i] data transfer direction  */
    int status;		
    struct scatterlist *sclp;	/* for struct scatterlist */
    /*Maintainence stuff*/
    Sgat_dio_hdr_t *u_io_hdrp;  
    Sgat_dio_hdr_t *k_io_hdrp;  
    void *schp;
} Sgat_dio_request;

///////////////////////////////////////////////////////////////////////////////
//
//API
//
//////////////////////////////////////////////////////////////////////////////
void sgat_dio_done( Sgat_dio_request *srp);
int sgat_dio_build_sg_list(Sgat_dio_hdr_t *io_hdrp, Sgat_dio_request *srp);
Sgat_dio_request *alloc_sgat_dio_request(void);
void free_sgat_dio_request(Sgat_dio_request *srp);
void sgat_dio_done_nocopy( Sgat_dio_request *srp);

#endif /* __KERNEL__ */

#endif /*_SGAT_DIO_H*/
